//---------------------------------------------------------------------------
//JPEG Loader
//Source by Johnny Wood
//
//This program uses the Intel JPEG Library to load JPEG images and
//copy them to Direct Draw surfaces.
//---------------------------------------------------------------------------
#include <windows.h>
#include <windowsx.h>
#include "ddraw.h"
#include "ddutil.h"
#include "resource.h"
#include "ijl.h"

#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif

#define NAME  "JPEG Loader by Johnny Wood"
#define TITLE "JPEG Loader by Johnny Wood"

//Display Mode
#define CLRMODE        24  //bit size
#define SCREENWIDTH    640 //horizontal screeen pixel count
#define SCREENHEIGHT   480 //vertical screen pixel count

//16Bit Color conversion macros
#define RGB555(r, g, b) ((r >> 3) | ((g >> 3) << 5) | ((b >> 3) << 10))
#define RGB565(r, g, b) ((r >> 3) | ((g >> 2) << 5) | ((b >> 3) << 11))

//16Bit Color Formats
#define PF_555  1
#define PF_565  2

//Color modes
#define C16BIT  16
#define C24BIT  24
#define C32BIT  32

//Where to get JPEG
#define FROM_FILE     0
#define FROM_RESOURCE 1

//****************
//Global variables
//****************

//Handles
HWND      hwnd;        //Handle to main window
HDC       hdc;         //Handle to a device context
HRESULT   Result;      //Results from function calls
HINSTANCE hinstance;   //Handle to program instance

//Direct Draw Object and Surfaces
LPDIRECTDRAW7		   DDObject       = NULL; // DirectDraw object
LPDIRECTDRAWSURFACE7	PrimarySurface = NULL; // DirectDraw primary surface
LPDIRECTDRAWSURFACE7	BackBuffer     = NULL; // DirectDraw back surface
LPDIRECTDRAWSURFACE7 JPGGraphic     = NULL; // JPG surface

//Direct Draw Blit Effects Object
DDBLTFX ddbltfx;

//Is application active?
bool Active = false; //Is application active?

//Graphics coordinates holder
RECT Image;

//16Bit format for video card
int  PixelFormat = PF_555;

//Function declarations
void DrawScreen(void);     //draws mouse, stamper, and buttons
bool Init(HINSTANCE, int); //Windows, DirectX Initialization
bool InitFail(HWND);       //Reports error and exits
bool LoadJPEG(int, LPCTSTR, LPDIRECTDRAWSURFACE7*);//Loads JPEG to DD Surface
void ProcessInput(void);   //keyboard input
void ReleaseObjects(void); //Clean up
HRESULT RestoreAll(void);  //Recover from ALT+TAB, etc.
LRESULT CALLBACK WindowProc(HWND hWnd, UINT message,
   WPARAM wParam, LPARAM lParam); //Windows Callback Procedure



//***************
//* Draw Screen *
//***************
void DrawScreen (void)
{
   //Clear old screen away
   ZeroMemory( &ddbltfx, sizeof(ddbltfx) );
   ddbltfx.dwSize = sizeof(ddbltfx);
   ddbltfx.dwFillColor = RGB(0,0,0); //Set to black
   BackBuffer->Blt(NULL, NULL, NULL, DDBLT_COLORFILL | DDBLT_WAIT, &ddbltfx);
      
   DDSURFACEDESC2 ddsd;
   ZeroMemory(&ddsd, sizeof(ddsd));
   ddsd.dwSize = sizeof( ddsd );

   //Get the surface description so that we can dynamically
   //find the width and height of our surface
   Result = JPGGraphic->GetSurfaceDesc(&ddsd);
   if (Result == DD_OK)
   {
      //Coordinates of image size
      Image.left   = 0;
      Image.top    = 0;
      Image.right  = ddsd.dwWidth;
      Image.bottom = ddsd.dwHeight;
   
      //Blit image to back buffer
      while (true)
      {
         Result = BackBuffer->BltFast (0, 0, JPGGraphic, &Image,
            DDBLTFAST_WAIT | DDBLTFAST_NOCOLORKEY);
         if( Result == DD_OK )  break;
         if( Result == DDERR_SURFACELOST )
         {  Result = RestoreAll();
            if( Result != DD_OK ) break; }
         if( Result != DDERR_WASSTILLDRAWING ) break;
      }
   }

   //Flip back buffer to the screen
   while(true)
   {  Result = PrimarySurface->Flip(NULL, DDFLIP_WAIT);
      if( Result == DD_OK )  break;
	   if( Result == DDERR_SURFACELOST )
	   {  Result = RestoreAll();
	      if( Result != DD_OK ) break; }
      if( Result != DDERR_WASSTILLDRAWING ) break;
   }

   return;
}



//**************
//* Initialize *
//**************
bool Init( HINSTANCE hInstance, int nCmdShow )
{
   WNDCLASSEX		wc;
   DDSURFACEDESC2 ddsd;

   //Set up Window class
   wc.style         = CS_HREDRAW | CS_VREDRAW;
   wc.lpfnWndProc   = WindowProc;
   wc.cbClsExtra    = 0;
   wc.cbWndExtra    = 0;
   wc.hInstance     = hInstance;
   wc.hIcon         = LoadIcon(hInstance, "LARGE");
   wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
   wc.hbrBackground = GetStockBrush(BLACK_BRUSH);
   wc.lpszMenuName  = NAME;
   wc.lpszClassName = TITLE;
   wc.cbSize        = sizeof(WNDCLASSEX);
   wc.hIconSm       = LoadIcon(hInstance, "SMALL");
   
   //Register Window class
   if (!RegisterClassEx(&wc))
   {   
      MessageBox( hwnd, "Failed to register window class", TITLE, MB_OK);
      return InitFail(hwnd);
   }

   //Create a window
   hwnd = CreateWindowEx(WS_EX_TOPMOST, NAME, TITLE, WS_VISIBLE | WS_POPUP,
      0, 0, GetSystemMetrics(SM_CXSCREEN), GetSystemMetrics(SM_CYSCREEN),
      NULL, NULL, hInstance, NULL);
   if(!hwnd)
   {
      MessageBox( hwnd, "Main window not created", TITLE, MB_OK);
      return InitFail(hwnd);
   }

	//Create DirectDraw object
   DirectDrawCreateEx(NULL, (void **)&DDObject, IID_IDirectDraw7, NULL );
   if(Result != DD_OK)
   {
      MessageBox(hwnd, "The DirectDraw Object could not be created",
         TITLE, MB_OK);
      return InitFail(hwnd);
   }

   //Get exclusive mode
   Result = DDObject->SetCooperativeLevel( hwnd,
      DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN);
   if(Result != DD_OK)
   {  
      MessageBox(hwnd, "Could not gain exclusive, fullscreen access",
         TITLE, MB_OK);
	   return InitFail(hwnd);
   }

   //Set the video mode
   Result = DDObject->SetDisplayMode(SCREENWIDTH,
      SCREENHEIGHT, CLRMODE, 0, 0);
   if(Result != DD_OK)
   {
      MessageBox(hwnd, "Set Display Mode failed", TITLE, MB_OK);
      return InitFail(hwnd);
   }

   //Prepare to create the primary surface with 1 back buffer
   ZeroMemory(&ddsd, sizeof(ddsd));
   ddsd.dwSize            = sizeof( ddsd );
   ddsd.dwFlags           = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
   ddsd.ddsCaps.dwCaps    = DDSCAPS_PRIMARYSURFACE | DDSCAPS_FLIP |
			                   DDSCAPS_COMPLEX;
   ddsd.dwBackBufferCount = 1;

   //Create our PrimarySurface
   Result = DDObject->CreateSurface( &ddsd, &PrimarySurface, NULL );
   if(Result != DD_OK)
   {
      MessageBox(hwnd, "The Primary Surface could not be created",
         TITLE, MB_OK);
      return InitFail(hwnd);
   }

   //Clear ddsd
   ZeroMemory(&ddsd, sizeof(ddsd));
   ddsd.dwSize = sizeof(ddsd);

   //Create the BackBuffer
   ddsd.ddsCaps.dwCaps = DDSCAPS_BACKBUFFER;
   Result = PrimarySurface->GetAttachedSurface(&ddsd.ddsCaps, &BackBuffer);
   if(Result != DD_OK)
   {  
      MessageBox(hwnd, "The Back Buffer could not be created",
         TITLE, MB_OK);
      return InitFail(hwnd);
   }

   //If we are in 16Bit color mode, find out the pixel format
   //that the video card uses
   if (CLRMODE == C16BIT)
   {
      DDPIXELFORMAT DDPixelFormat;
      ZeroMemory(&DDPixelFormat, sizeof(DDPixelFormat));
      DDPixelFormat.dwSize = sizeof(DDPixelFormat);

      Result = PrimarySurface->GetPixelFormat(&DDPixelFormat);
      if(Result != DD_OK)
      {
         MessageBox(hwnd, "Failed to retrieve pixel format", TITLE, MB_OK);
         return InitFail(hwnd);
      }
      
      //Get green bitmask and set pixel format    
      if((DDPixelFormat.dwGBitMask / 32) == 31)
         PixelFormat = PF_555;
      else if((DDPixelFormat.dwGBitMask / 32) == 63)
         PixelFormat = PF_565;
   }

   //Example: Loading resource using a string ID
   Result = LoadJPEG(FROM_RESOURCE, "EXAMPLE", &JPGGraphic);
   
   //Example: Loading resource using an integer ID
   //The NULL string implies FROM_RESOURCE so we
   //just pass the EXAMPLE integer to the function
   //Result = LoadJPEG(EXAMPLE, NULL, &JPGGraphic);

   //Example: Loading from a file
   //Result = LoadJPEG(FROM_FILE, "EXAMPLE.JPG", &JPGGraphic);
   
   //Test success of LoadJpeg function
   if (Result == false)
   {
      MessageBox(hwnd, "Failed to load JPEG", TITLE, MB_OK);
      return InitFail(hwnd);
   }

   //Optional: Set the Color Key (Transparent Color)
   //DDSetColorKey(JPGGraphic, RGB(0,0,0));

   //DirectDraw initialized without fail
   return true;
}



//***************
//* Init Failed *
//***************
//Initialization failed
bool InitFail(HWND hwnd)
{
   ReleaseObjects();
   DestroyWindow(hwnd);
   return false;
}



//*************
//* Load JPEG *
//*************
bool LoadJPEG(int Location, LPCTSTR Name,  LPDIRECTDRAWSURFACE7* Surface)
{
   //Error handle
   IJLERR jerr;

   //Create the JPEG object
   JPEG_CORE_PROPERTIES jcprops;
   
   //Initialize the JPEG object
   jerr = ijlInit(&jcprops);
   if (jerr != IJL_OK)
   {
      MessageBox(hwnd, "IJLInit problem", TITLE, MB_OK);
      return false;
   }

   //Load from File or Resource?
   if (Location == FROM_FILE)
   {
      //Set the IJL data source as a filename
      jcprops.JPGFile = Name;

      //Read JPEG parameters from the file
      jerr = ijlRead(&jcprops, IJL_JFILE_READPARAMS);
   }
   else //(Location == FROM_RESOURCE) or integer ID of resource
   {
      //Resource variables
      BYTE* pmem;
      HGLOBAL hmem;
      DWORD size;
      HRSRC hres;
      
      //Determine whether resource ID is an int or a string
      if (Name == NULL)
         hres = FindResource(NULL, MAKEINTRESOURCE(Location), "JPEG");
      else
         hres = FindResource(NULL, "EXAMPLE", "JPEG");      
      if (hres)
      {
         size = SizeofResource(NULL, hres);
         if (!size)
         {
            MessageBox(hwnd, "Error retrieving resource size",
               TITLE, MB_OK);
            return false;
         }
         hmem = LoadResource(NULL, hres);
         if (hmem == NULL)
         {
            MessageBox(hwnd, "Error loading resource", TITLE, MB_OK);
            return false;
         }
         pmem = (BYTE *)LockResource(hmem);
         if (pmem == NULL)
         {
            MessageBox(hwnd, "Error locking resource", TITLE, MB_OK);
            return false;
         }
      }
      else
      {
         MessageBox(hwnd, "JPEG resource not found", TITLE, MB_OK);
         return false;
      }

      //Set the IJL data source as the resource buffer
      jcprops.JPGFile      = NULL;
      jcprops.JPGBytes     = pmem;
      jcprops.JPGSizeBytes = size;

      //Read JPEG parameters from the buffer
      jerr = ijlRead(&jcprops, IJL_JBUFF_READPARAMS);
   }
   
   //Make sure Parameter read was successful
   if (jerr != IJL_OK)
   {
      MessageBox(hwnd, "Error reading JPEG parameters", TITLE, MB_OK);
      return false;
   }

   //Prepare a 24Bit buffer to receive image data
   BYTE *buffer24;
   
   //Determine the required size
   long szbuff24 = (jcprops.JPGWidth * C24BIT + 7) / 8
      * jcprops.JPGHeight;

   //Resize the buffer
   buffer24 = new BYTE [szbuff24];
   if (buffer24 == NULL)
   {
      MessageBox(hwnd, "Memory Allocation Error", TITLE, MB_OK);
      return false;
   }
   
   //Set up the DIB specification for the JPEG decoder
   jcprops.DIBWidth    = jcprops.JPGWidth;
   jcprops.DIBHeight   = jcprops.JPGHeight; //Implies a bottom-up DIB.
   jcprops.DIBChannels = 3;
   jcprops.DIBColor    = IJL_BGR;
   jcprops.DIBPadBytes = IJL_DIB_PAD_BYTES(jcprops.JPGWidth,3);
   jcprops.DIBBytes    = reinterpret_cast<BYTE*>(buffer24);

   //Set the JPG color space ... this will always be somewhat of an
   //educated guess at best because JPEG is "color blind" (i.e.,
   //nothing in the bit stream tells you what color space the data was
   //encoded from. However, in this example we assume that we are
   //reading JFIF files which means that 3 channel images are in the
   //YCbCr color space and 1 channel images are in the Y color space.
   switch(jcprops.JPGChannels)
   {
      case 1: jcprops.JPGColor = IJL_G;
         break;
      
      case 3: jcprops.JPGColor = IJL_YCBCR;
         break;
      
      default:
         //This catches everything else, but no color twist will be
         //performed by the IJL.
         jcprops.DIBColor = (IJL_COLOR)IJL_OTHER;
         jcprops.JPGColor = (IJL_COLOR)IJL_OTHER;
         break;
   }
   
   //Read JPEG image data into our 24bit buffer
   if (Location == FROM_FILE)
      jerr = ijlRead(&jcprops, IJL_JFILE_READWHOLEIMAGE);
   else
      jerr = ijlRead(&jcprops, IJL_JBUFF_READWHOLEIMAGE);

   //Make sure the read was successful
   if (jerr != IJL_OK)
   {
      MessageBox(hwnd, "Error reading JPEG image", TITLE, MB_OK);
      return false;
   }

   HBITMAP hbm;

   //Convert to current CLRMODE
   if (CLRMODE == C16BIT)
   {
      //Create a 16bit buffer
      WORD *buffer16;
      long szbuff16;
      
      //determine the size of our buffer
      szbuff16 = ((jcprops.JPGWidth * C16BIT + 7) / 8)
         * jcprops.JPGHeight;
      
      //resize the buffer and make sure resize works
      buffer16 = new WORD [szbuff16];
      if (buffer16 == NULL)
      {
         MessageBox(hwnd, "Error creating 16Bit buffer", TITLE, MB_OK);
         return false;
      }

      //Start at the beginning of the buffer
      long j = 0;

      //Step through the 24bit buffer
      //Retrieve 3 channels at a time and convert their values to 16bit
      for (long i = 0; i < szbuff24; i += 3)   
      {
         //Check the pixel format and write the color data
         //to the 16bit buffer. After the write, advance the
         //16bit buffer by one.
         if (PixelFormat == PF_555)
            buffer16[j++] = RGB555(buffer24[i], buffer24[i + 1],
               buffer24[i + 2]);
         else
            buffer16[j++] = RGB565(buffer24[i], buffer24[i + 1],
               buffer24[i + 2]);
      }

      //Create the bitmap using the new 16bit buffer   
      hbm = CreateBitmap (jcprops.JPGWidth, jcprops.JPGHeight, 1,
         C16BIT, buffer16);
      if(hbm == NULL)
      {
         MessageBox(hwnd, "Failed to create 16Bit Bitmap", TITLE, MB_OK);
         return false;
      }

      //remove the new buffer
      delete buffer16;
   }
   else if (CLRMODE == C24BIT)
   {
      //The data we have from the JPEG is already 24bit
      //Just create the new bitmap from our buffer
      hbm = CreateBitmap (jcprops.JPGWidth, jcprops.JPGHeight, 1,
         C24BIT, buffer24);
      if(hbm == NULL)
      {
         MessageBox(hwnd, "Failed to create 24Bit Bitmap", TITLE, MB_OK);
         return false;
      }
   }
   else if (CLRMODE == C32BIT)
   {
      //Create a 32bit buffer
      BYTE *buffer32;
      long szbuff32;
      
      //determine the size of our buffer
      szbuff32 = ((jcprops.JPGWidth * C32BIT + 7) / 8)
         * jcprops.JPGHeight;

      //resize the buffer and make sure resize works
      buffer32 = new BYTE [szbuff32];
      if (buffer32 == NULL)
      {
         MessageBox(hwnd, "Error creating 32Bit buffer", TITLE, MB_OK);
         return false;
      }
      
      //Start at the beginning of the 24bit buffer
      long j = 0;

      //Step through the 32bit buffer, copying the 3 Channels from
      //the 24bit buffer. However, at the end of each channel, we
      //have to write an extra byte for the 32bit image's alpha
      //channel
      for (long i = 0; i < szbuff32; i += 4)   
      {
         buffer32[i + 0] = buffer24[j++]; //Blue channel
         buffer32[i + 1] = buffer24[j++]; //Green channel
         buffer32[i + 2] = buffer24[j++]; //Red channel
         buffer32[i + 3] = (BYTE)0x00;    //Alpha channel
      }

      //Create the bitmap using the new 32bit buffer   
      hbm = CreateBitmap (jcprops.JPGWidth, jcprops.JPGHeight, 1,
         C32BIT, buffer32);
      if(hbm == NULL)
      {
         MessageBox(hwnd, "Failed to create 32Bit Bitmap",
            TITLE, MB_OK);
         return false;
      }

      //remove the 32bit buffer
      delete buffer32;
   }
   else
   {
      MessageBox(hwnd,
         "Palettized modes not supported. Use a 16, 24, or 32 Bit display.",
         TITLE, MB_OK);
      return false;
   }

   //remove the old buffer
   delete buffer24;

   //Create a DirectDraw surface   
   DDSURFACEDESC2 ddsd;

   //If DirectDraw Surface is NULL, create it
   if (*Surface == NULL)
   {
      // Create a DirectDrawSurface for this bitmap
      ZeroMemory(&ddsd, sizeof(ddsd));
      ddsd.dwSize   = sizeof(ddsd);
      ddsd.dwFlags  = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
      ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
      ddsd.dwWidth  = jcprops.JPGWidth;
      ddsd.dwHeight = jcprops.JPGHeight;

      //Create the surface from our surface description
      Result = DDObject->CreateSurface(&ddsd, Surface, NULL);
      if(Result != DD_OK)
      {
         MessageBox(hwnd, "Could not create surface", TITLE, MB_OK);
         return false;
      }
   }

   //Copy bitmap data to DirectDraw Surface
   DDCopyBitmap(*Surface, hbm, 0, 0, 0, 0);
      
   // clean up and destroy the JPEG Decompressor
   ijlFree(&jcprops);

   return true;
}



//*****************
//* Process Input *
//*****************
void ProcessInput(void)
{
   //Check for keyboard input
   if (HIBYTE(GetAsyncKeyState(VK_ESCAPE)))
      PostMessage(hwnd, WM_CLOSE, 0, 0);
   else if (HIBYTE(GetAsyncKeyState(VK_RETURN)))
      PostMessage(hwnd, WM_CLOSE, 0, 0);
   else if (HIBYTE(GetAsyncKeyState(VK_SPACE)))
      PostMessage(hwnd, WM_CLOSE, 0, 0);

   return;
}



//*******************
//* Release Objects *
//*******************
static void ReleaseObjects(void)
{
   if(DDObject != NULL)
   {
	   if(PrimarySurface != NULL)
	   {
         PrimarySurface->Release();
	      PrimarySurface = NULL;
	   }
      if(JPGGraphic != NULL)
	   {
         JPGGraphic->Release();
	      JPGGraphic = NULL;
	   }
      
      DDObject->Release();
	   DDObject = NULL;
   }
}



//*******************
//* Restore Objects *
//*******************
HRESULT RestoreAll(void)
{
   Result = PrimarySurface->Restore();
   if (Result == DD_OK)
   {
      Result = JPGGraphic->Restore();
      if (Result == DD_OK)
        LoadJPEG(FROM_RESOURCE, "SAMPLE", &JPGGraphic);
        //LoadJPEG(SAMPLE, NULL, &JPGGraphic);
        //LoadJPEG(FROM_FILE, "EXAMPLE.JPEG", &JPGGraphic);
   }
   return Result;
}



//********************
//* Window Procedure *
//********************
LRESULT CALLBACK WindowProc(HWND hWnd, UINT message,
   WPARAM wParam, LPARAM lParam)
{
   switch(message)
   {
      case WM_CREATE:
         break;

      //This determines if we are the active application
      case WM_ACTIVATEAPP:
         // Pause if minimized or not the top window
         Active = (wParam == WA_ACTIVE) || (wParam == WA_CLICKACTIVE);
         break;

      //We are quitting the application
      case WM_DESTROY:
    	   ReleaseObjects();
	      PostQuitMessage(0);
         break;

      default:
         return DefWindowProc(hWnd, message, wParam, lParam);
   }
   return(0L);
}



//************
//* Win Main *
//************
int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
   LPSTR lpCmdLine, int nCmdShow)
{
   MSG msg;

   lpCmdLine     = lpCmdLine;
   hPrevInstance = hPrevInstance;
   hinstance     = hInstance;

   //Initialize Direct Draw
   if(!Init(hInstance, nCmdShow))
   	return false;

   //main loop
   while(true)
   {
      //Only do these things if application is active
	   if (Active)
      {
         DrawScreen();
         ProcessInput();
      }

      //See if there are any Windows messages, process them
      if (PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE))
      {
  	     if (!GetMessage(&msg, NULL, 0, 0))
         	return msg.wParam;
        TranslateMessage(&msg);
        DispatchMessage(&msg);
      }
   }

   return EXIT_SUCCESS;
}
